/**
 * Author: Hendrik van Arragon, 2021
 * SPDX-License-Identifier: MIT
 */

#ifndef UTIL_INC_CLOCK_H_
#define UTIL_INC_CLOCK_H_

#include <cstdint>

namespace abmt{

class time{
public:

	int64_t ns_since_1970; // damit negative deltas leichter abbildbar sind

	time(int64_t s=0):ns_since_1970(s){

	}

	/// must be defined in runtime.
	static time now();

	static time ns(int64_t t){
		return time(t);
	}

	static time us(int64_t t){
		return time(t * 1000);
	}

	static time ms(int64_t t){
		return time(t * 1000 * 1000);
	}

	static time sec(int64_t t){
		return time(t * 1000 * 1000 * 1000);
	}

	static time min(int64_t t){
		return time(t * 1000 * 1000 * 1000 * 60);
	}

	static time hour(int64_t t){
		return time(t * 1000 * 1000 * 1000 * 60 * 60);
	}

	static time day(int64_t t){
		return time(t * 1000 * 1000 * 1000 * 60 * 60 * 24);
	}

	static time week(int64_t t){
		return time(t * 1000 * 1000 * 1000 * 60 * 60 * 24 * 7);
	}

	static time year(int64_t t){
		return time(t * 1000 * 1000 * 1000 * 60 * 60 * 24 * 7 * 52);
	}


	double ns(){
		return (double) ns_since_1970;
	}

	double us(){
		return (double) ns_since_1970/(1000);
	}

	double ms(){
		return (double) ns_since_1970/(1000*1000);
	}

	double sec(){
		return (double) ns_since_1970/(1000*1000*1000);
	}

	double min(){
		return (double) ns_since_1970/(1000LL*1000*1000*60);
	}

	double hour(){
		return (double) ns_since_1970/(1000LL*1000*1000*60*24);
	}

	double week(){
		return (double) ns_since_1970/(1000LL*1000*1000*60*24*7);
	}

	double year(){
		return (double) ns_since_1970/(1000LL*1000*1000*60*24*7*52);
	}

	bool operator==(const time &t) { return ns_since_1970 == t.ns_since_1970; }
	bool operator!=(const time &t) { return ns_since_1970 != t.ns_since_1970; }
	bool operator< (const time &t) { return ns_since_1970 <  t.ns_since_1970; }
	bool operator> (const time &t) { return ns_since_1970 >  t.ns_since_1970; }
	bool operator<=(const time &t) { return ns_since_1970 <= t.ns_since_1970; }
	bool operator>=(const time &t) { return ns_since_1970 >= t.ns_since_1970; }

	time& operator+=(const time &t){
		ns_since_1970 += t.ns_since_1970;
		return *this;
	}

	time& operator-=(const time &t){
		ns_since_1970 -= t.ns_since_1970;
		return *this;
	}

	time& operator*=(const time &t){
		ns_since_1970 *= t.ns_since_1970;
		return *this;
	}

	time& operator/=(const time &t){
		ns_since_1970 /= t.ns_since_1970;
		return *this;
	}

	time operator+(const time &t) const{
		time res(*this);
		res += t;
		return res;
	}
	time operator-(const time &t) const{
		time res(*this);
		res -= t;
		return res;
	}
	time operator*(const time &t) const{
		time res(*this);
		res *= t;
		return res;
	}
	time operator/(const time &t) const{
		time res(*this);
		res /= t;
		return res;
	}

	time operator*(const double& n) const{
		time res(*this);
		res.ns_since_1970 *= n;
		return res;
	}

	time operator/(const double& n){
		time res(*this);
		res.ns_since_1970 /= n;
		return res;
	}

};

// for things like time_obj*1.5
time operator*(const double& n,const time &rhs);

} // namespace

#endif /* UTIL_INC_CLOCK_H_ */
